#!/usr/bin/env perl -w
#
# wikibackups - backup wikis
#
# Author: Tamara Temple <tamara@tamaratemple.com>
# Created: 2011/10/23
# Time-stamp: <2012-07-11 00:05:45 tamara>
# Copyright (c) 2011 Tamara Temple Web Development
# License: GPLv3
#

use strict;
use YAML::Syck;
use Data::Dumper;
use Getopt::Std;

my $VERSION = '0.1';

our($opt_c,$opt_n,$opt_d,$opt_v,$opt_V,$opt_h);
usage() unless getopts('c:vndVh');
usage() if $opt_h;
version() if $opt_V;

my $debug=$opt_d;
my $verbose=$opt_v;
my $dryrun=($opt_n?"--dry-run":"");
my $configfilename=($opt_c?$opt_c:$ENV{'HOME'}.'/.wbu.rc');
if (! -r $configfilename) {
    print("No rc file found!");
    usage();    
}
my $config_ref = LoadFile($configfilename);

my $rsync_cmd = '/usr/bin/rsync';

my $exclusions_file="/tmp/exc.$$";

if (! $config_ref) {
    die("Errors loading rc file!");
}

my %config = %$config_ref;

warn "\%config=".Dumper(\%config) if $debug;

foreach my $wiki (keys %config) {
    backup_wiki($config{$wiki},$wiki);
}

sub backup_wiki {
    my $wiki_ref = shift @_;
    my $wikiname = shift @_;
    my %wiki = %$wiki_ref;
    print "Backing up $wikiname\n" if $verbose;
    my $exclfilename = save_exclusions($wiki{'exclusions'},$wikiname);

    my $SOURCE = $wiki{'source'};
    my $TARGET = $wiki{'target'};

    my $LINKDEST = get_last_update($TARGET);
    warn "from get_last_update: \$LINKDEST = $LINKDEST\n" if $debug;
    $LINKDEST = ($LINKDEST eq ''?'':"--link-dest=$LINKDEST");
    warn "after making rsync option: \$LINKDEST = $LINKDEST\n" if $debug;
    

    if ((! -d $TARGET) && (! -w $TARGET) && (! -x $TARGET)) {
	die("$TARGET not usable");
    }


    # Check if an incomplete backup already exists
    # If so, rename it
    if ( -e "$TARGET/.incomplete" ) {
	my $save_incomplete = "$TARGET/.incomplete.".datestring();
	warn "$TARGET/.incomplete already exists, saving it in $save_incomplete" if $verbose;
	rename("$TARGET/.incomplete",$save_incomplete);
    }

    my @rsyncoptions = (
	$dryrun,                # only do a dry-run, don't transfer anything
	($verbose?"--verbose --progress --itemize-changes":""), # talk loudly about what is happening
	"--archive",		# make and archive, equivalent to rlptgoD options
	"--compress",		# compress file data on transfer
	"--one-file-system",	# don't cross file system boundaries
	"--hard-links",		# preserve hard links
	"--inplace",		# update destination file in place (do not create a new copy)
	"--numeric-ids",	# don't map uid/gid values by user/group
	"--delete",		# remove deleted files
	$LINKDEST,              # destination directory for forming hard links from
	"--exclude-from=$exclfilename", # file of exclusions
	"$SOURCE",		# source to pull from
	"$TARGET/.incomplete",	# target to write to
	"2>&1"			# send STDERR to STDOUT
	);

    warn "\@rsyncoptions: ".Dumper(\@rsyncoptions) if $debug;

    my $cmd=$rsync_cmd . join(" ",@rsyncoptions);
    warn "\$cmd=$cmd" if $debug;
    print "$cmd\n" if $verbose;
    # return if $debug;
    my @result = qx{$cmd};
    die("rsync failed: $!") if $! > 0;
    print "Rsync output:\n" . join("",@result) . "\n\n" if $verbose;

    if (-e "$TARGET/.incomplete") {

	# Rename .incomplete and save in $TARGET/.lastbackup
	my $backupname = datestring();
	rename("$TARGET/.incomplete","$TARGET/$backupname");
	
	save_last_update($TARGET,"$TARGET/$backupname");
    } else {
	print "No backup made.\n" if $verbose;
    }
}

sub save_exclusions
{
    my ($excl_ref, $wikiname) = @_;

    my @excl = @$excl_ref;
    warn Dumper(@excl) if $debug;
    @excl = map { (convert($_)) } @excl;
    warn Dumper(@excl) if $debug;

    my $exclfile = "$exclusions_file.$wikiname";

    open(FH,">","$exclfile") or die("Could not open $exclfile");
    print FH join("\n",@excl)."\n";
    close FH;
    print "\nExclusions:\n".join("\n",@excl)."\n\n" if $verbose;
    return $exclfile;
}

sub convert {
    my $s = shift @_;
    $s =~ s/^i /+ /;
    $s =~ s/^e /- /;
    warn "in convert: \$s=$s\n" if $debug;
    return $s;
}


sub get_last_update {
    my $target = shift @_;
    my $text;
    if (-e "$target/.lastupdate") {
	$text = do { local( @ARGV, $/ ) = "$target/.lastupdate" ; <> } ;
	warn "\$text: $text" if $debug;
	chomp($text);
    } else {
	warn "$target/.lastupdate was not found" if $debug;
	$text = '';
    }
    return $text;
}

sub save_last_update {
    my ($target, $filename) = @_;
    open(FH,">","$target/.lastupdate");
    print FH $filename;
    close(FH);
}

sub datestring {
    my $ds = qx{date "+%Y-%m-%dT%H:%M:%S"}; chomp($ds);
    return $ds;
}


sub version {
    print "$0 version: $VERSION\n";
    exit;
}

sub usage {
    print "$0 [-c config] [-n] [-d] [-v] [-V]\n";
    print "\n";
    print "where:\n";
    print "   -c - specify an alternate configuration file.\n";
    print "        default is \$HOME/.wbu.rc\n";
    print "   -n - do a dry run only, no files will be transfered.\n";
    print "        Most useful with -d option to see output of rsync command\n";
    print "   -d - turns on debug messages\n";
    print "   -v - turns on verbose mode\n";
    print "   -V - shows version\n";
    print "   -h or -? - gives this help message\n";
    print "\n";
    exit;
}


# ****************
# Documentation!!!
# ****************
    
=pod

=head1 NAME

=over 2

wikibackups.pl

=back

=head1 SYNOPSIS

=over 4

wikibackups.pl B<[-c config]> B<[-n]> B<[-v]> B<[-d]>

wikibackups.pl B<[-h]>

wikibackups.pl B<[-V]>

=back

=head1 DISCUSSION

=over 2

B<wikibackups.pl> provides a way of automating backups of L<http://www.pmwiki.org/> wikis. This script uses the C<rsync(1)> to handle the actual file transfer. Various inclusions and exclusion can be set (see L<"CONFIGURATION">).

More discussion on backing up your wiki can be found at L<http://www.pmwiki.org/wiki/PmWiki/BackupAndRestore>.

=back 

=head1 OPTIONS

=over 4

B<-c config> - specify an alternate configuration file. The default is C<$HOME/.wbu.rc>.

B<-n> - dry run only. If used with B<-v> will show the list of files which rsync would have transfered.

B<-v> - verbose mode. Will print out what is going on during the backup, including the output from the rsync command.

B<-d> - debug mode. Will print out all sorts of internal structure information. Useful for checking your configuration file. If you want to check your configuration, make sure to set -n and -v as well.

B<-h> - help. Prints a usage message.

B<-V> - version. Prints the script version.

=back

=head1 CONFIGURATION

=over 2

Configuration of wikibackup.pl is handled with a YAML file. The default location is in C<$HOME/.wbu.rc> but can be specified using the C<-c> parameter on the command line.

YAML (see L<http://yaml.org/>) is a handy language to set up nested configurations with. It generally has the syntax of:

   item1: value1
   item2: value2
     item2.1: value2.1
     item2.2: value2.2
       item2.2.1: value2.2.1
   item2.3: value2.3

And so on.

The C<.wbu.rc> file is configured thusly:

   wikiname:
      source: user@host:path/to/wikiroot
      target: path/to/backuproot
   # Inclusions are marked by "i"
   # Exclusions are marked by "e"
      exclusions:
          - i /cookbook/
          - i /wiki.d/
          - e /wiki.d/.flock
          - e /wiki.d/.pageindex
          - e /wiki.d/.lastmod
          - e /wiki.d/*,del-*
          - e /wiki.d/*/*,del-*
          - i /pub/
          - i /local/
          - i /uploads/
          - e /.git/*
          - e /*
          - e **~
          - e **.bak
          - e **.tgz
          - e **.zip
          - e **.gz
          - e **.Z

Most of the entries should make sense if you're familiar with the structure of the wiki's directories.

A breakdown of the configuration follows:

=back

=over 6

B<wikiname> - this is the name you give your particular wiki. It doesn't have to be the same as C<$WikiTitle>, but it's helpful if it at least resembles that. This value will be used as the backup directory underneath the path set in backuppath.

B<source> - this is the rsync(1) spec path for the source of the files. (There are actually few restrictions on this spec as to whether it is local or remote. The assumption is that it will be a remote server that one has C<ssh> access to.

B<target> - this is the rsync(2) spec path for the root directory of the backup tree.

B<exclusions> - here is where you list what to include and exclude from your backup. See L<"Configuring Exclusions">.

=back

=over 2

Lines that begin with an octothorpe (#) are treated a comments. 

=back

=head2 Warning about TABS in YAML input

YAML chokes on tab characters in it's input. Make sure to always use spaces to perform indenting. (Emacs has an C<untabify> function that's particularly useful for this.)

=head2 Configuring Exclusions

=over 2

This may be the most difficult part of this. C<rsync>'s method of filtering what gets sent and what doesn't is pretty arcane. Some rules of thumb, though:

=back

=over 6

=item *

Inclusions must precede Exclusions.

=item * 

If an item begins with a directory separator (/), then it is considered only at the root of the rsync path.

=item *

If an item ends with a directory separator (/) then it is treated as a directory and it's inclusion or exclusion applies to it and everything below.

=item *

Normal shell globbing is used, so C<something*> will match everything that begins with "something":

  something
  something.png
  somethingelse.txt
  something.nice.for.you.try

=item *

In the configuration file, inclusions are marked by B<i> in the configuration file -- this is converted to a B<+> for rsync's filter file. Likewise, exclusions are marked by B<e> in the configuration file -- these are converted to B<-> for rsync's filter file.

=back

=head1 EXAMPLES

=over 2

This will perform a backup of B<mywiki>, on server B<wiki.example.com>, user B<me>, with the path to mywiki in the directory B<public_html/wiki> in the user's login area. The backup will be saved in rolling directories in B</mnt/backups/WikiBackups/mywiki>:

   mywiki:
      source: me@mywiki.example.com/public_html/wiki
      target: /mnt/backups/WikiBackups/mywiki
      exclusions:
        - i /cookbook/
        - i /wiki.d/
        - e /wiki.d/.flock
        - e /wiki.d/.pageindex
        - e /wiki.d/.lastmod
        - e /wiki.d/*,del-*
        - e /wiki.d/*/*,del-*
        - i /pub/
        - i /local/
        - i /uploads/
        - e /*

The exclusions section is as follows:

=back

=over 6

B<i /cookbook/> - include the cookbook directory, but only from the top level. This will include the entire contents of the C<cookbook> directory, but it won't include any occurances of C<cookbook> in other root directories lower down.

B<i /wiki.d/> - include the C<wiki.d> directory, but again, only from the top level.

B<e /wiki.d/.flock> - exclude the .flock file that pmwiki uses to control access. This is a generated file and should not be backed up or restored.

B<e /wiki.d/.pageindex> - exclude the .pageindex file that pmwiki creates to index wiki pages. This is also a generated file and should not be backed up or restored.


B<e /wiki.d/.lastmod> - exclude the .lastmod file that pmwiki creates. This is generated as well, and should not be backed up or restored.

B<e /wiki.d/*,del-*> - this is the pattern used by pmwiki to denote deleted wiki pages. If you do want to keep these in a backup, remove this line.

B<e /wiki.d/*/*,del-*> - this is similar to the above, but works when your pages are stored in group directories.

B<i /pub/> - the public directory for skins, themes, buttons, images, etc that you use to customize your wiki.

B<i /local/> - the local directory where you store you local configuration files.

B<i /uploads/> - uploads directory where attachments to pages are stored. Definitely want to keep this backed up along with the pages themselves.

B<e /*> - this is the last entry in this set -- it is telling rsync to exclude everything not explicitly included in the copy. This is the magic that makes it all work.

=back

=head2 Some additional helpful exclusions

=over 2

Sometimes, when working on a wiki installation, there may be cruft that gets left behind that you don't really want in your backups. Editor and tool backup files, downloaded archive and compressed files, etc. To deal with these, rsync looks at a double C<**> and says "match anything, regardless of directory". The single C<*> will match only within a directory (between directory separators (C</>)). The double star then makes it convenient for matching file types anywhere. Here are some examples:

          - e **~
          - e **.bak
          - e **.tgz
          - e **.zip
          - e **.gz
          - e **.Z


=back

=over 6

B<e **~> - typically editor backup files, especially C<Emacs>.

B<e **.bak> - typically utility or filter backup files, such as from C<sed> or C<perl>

B<e **.tgz> - a tarball (gzipped tar file).

B<e **.zip> - a zip compressed file.

B<e **.gz> - a gzip compressed file.

B<e **.Z> - another type of compressed file.

=back

=over 2

There are probably other kinds of files that you may have laying around that you don't really want to include in a site backup.

=back

=head1 PERIODIC BACKUPS

=over 2

This script performs a rolling forward backup as described in L<http://www.sanitarium.net/golug/rsync_backups_2010.html>. A new timestamped directory is created each time the script is run, linking files that have not changed to the last backup, and only copying or deleting files that have changed on the source.

Before scheduling this script automatically, make sure it works! Test the configuration using the -d, -v, and -n switches. When you are satisfied with how it is set up, then run it live once with -v turned on and make sure that you do get what you think you should. PmWikis' data doesn't tend to be very large, unless you have a lot of media stored as attachments. (If that is the case, you may consider another backup solution for your media and add the uploads directory to the exclusions list.)

Once you're satisfied everything is working correctly, you can add the script, without paramters, to your likely cron-type queue (cron, anacron, periodic, launchd, etc.). Set it to go off on a schedule for backing up that you feel comfortable with. In a fairly active wiki, daily should be the minimum. If you're wiki doesn't change much week-to-week or month-to-month, longer backup periods can be justified. It's entirely up to you. For most people, daily will be about right.

=back

=head2 Backup Directory Structure

=over 4

The rolling backups are stored underneath the B<target:> specified in the configuration file. Each back is run, the directory for running the rsync is called C<.incomplete>. When rsync ends without an error, the C<incomplete> directory is renamed with a timestamp. The path of that directory specified by C<TARGET/TIMESTAMP> is stored in the C<TARGET> directory in the C<.lastbackup> file.

Each periodic backup only holds the information from your wiki that actually changed that iteration. Files which are in previous backups unchanged are simply hard linked into the directory, thus saving space if nothing changes from backup to backup.

   TARGET/
      .lastbackup
      2012-07-09T14.28.04/
         (backed up files and directories)
      2012-07-10T14.30.33/
         (linked and changed files and directories)

=back

=head1 OTHER CONSIDERATIONS

=over 2

C<rsync> relies heavily on C<ssh>, which provides encrypted end-to-end communications between machines. As such, you should have C<ssh> access to any remote machine in this setup. Make sure the remote side already has the ssh public key from the local machine so there won't be any hiccups with authentication or having to supply a password to get C<rsync> to work correctly.

There are other ways of connecting to the remote side in C<rsync>. Such methods are beyond the scope of this project to describe.

=back

=head1 TODO

=over 6

=item *

Discuss how to restore a file from the rsync backup.

=back


=head1 SEE ALSO

=over 2

L<http://www.pmwiki.org/wiki/PmWiki/BackupAndRestore>, L<http://www.pmwiki.org/wiki/Cookbook/BackUpPages>, Thread (L<http://thread.gmane.org/gmane.comp.web.wiki.pmwiki.user/20317>) on the pmwiki-users group

=back

=head1 AUTHOR(S)

=over 2

Tamara Temple <tamara @ tamaratemple.com> L<http://www.pmwiki.org/wiki/Profiles/Tamouse>

=back

=cut

